﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Text;

namespace OsakaUniv.RIWL.Koudo.Writimation
{
    public class DrawingData
    {
        private const float ANCHORWIDTH = 8.0F;
        private const float ANCHORHEIGHT = 8.0F;
        private List<PointDataCollection> points_;
        private PointDataCollection anchorPoints_;
        private Bitmap backgroundImage_;
        private int penSize_;
        private Color penColor_;
        private float offsetX_;
        private float offsetY_;
        private float scale_;
        private string text_;
        private Font font_;
        private Color textColor_;

        #region constructor
        //
        //  constructor
        //

        public DrawingData()
        {
            this.points_ = new List<PointDataCollection>();
            this.anchorPoints_ = new PointDataCollection();
            this.Clear();
        }

        #endregion  //  constructor

        #region property
        //
        //  property
        //

        public PointDataCollection AnchorPoints
        {
            get { return this.anchorPoints_; }
            set { this.anchorPoints_ = value; }
        }

        public Bitmap BackgroundImage
        {
            get { return this.backgroundImage_; }
            set { this.backgroundImage_ = value; }
        }

        public Font Font
        {
            get { return this.font_; }
            set { this.font_ = value; }
        }

        public string FontName
        {
            get { return this.font_.Name; }
            set 
            {
                this.font_ = new Font(
                    value,
                    this.font_.Size,
                    this.font_.Style);
            }
        }

        public float FontSize
        {
            get { return this.font_.Size; }
            set
            {
                this.font_ = new Font(
                    this.font_.Name,
                    value,
                    this.font_.Style);
            }
        }

        public float OffsetX
        {
            get { return this.offsetX_; }
            set { this.offsetX_ = value; }
        }

        public float OffsetY
        {
            get { return this.offsetY_; }
            set { this.offsetY_ = value; }
        }

        public Color PenColor
        {
            get { return this.penColor_; }
            set { this.penColor_ = value; }
        }

        public int PenSize
        {
            get { return this.penSize_; }
            set { this.penSize_ = value; }
        }

        public List<PointDataCollection> Points
        {
            get { return this.points_; }
            set { this.points_ = value; }
        }

        public float Scale
        {
            get { return this.scale_; }
            set { this.scale_ = value; }
        }

        public string Text
        {
            get { return this.text_; }
            set { this.text_ = value; }
        }

        public Color TextColor
        {
            get { return this.textColor_; }
            set { this.textColor_ = value; }
        }

        #endregion  //  property

        #region method
        //
        //  method
        //

        public void Add(
            float x,
            float y,
            long time)
        {
            try
            {
                this.points_.Add(
                    new PointDataCollection(
                        this.NormalizePoint(x, y), 
                        time));
            }
            catch (Exception exp)
            {
                throw new Exception(
                    "[DrawData.Add(x, y, time)]"
                    + "(" + exp.Source + ")\r\n"
                    + "  " + exp.Message,
                    exp);
            }
        }

        public void AddPoint(
            int index,
            float x,
            float y,
            long time)
        {
            try
            {
                if (index < this.points_.Count)
                {
                    this.points_[index].Add(
                        this.NormalizePoint(x, y), 
                        time);
                }
            }
            catch (Exception exp)
            {
                throw new Exception(
                    "[DrawData.AddPoint(index, x, y, time)]"
                    + "(" + exp.Source + ")\r\n"
                    + "  " + exp.Message,
                    exp);
            }
        }

        public void AddPoint(
            float x,
            float y,
            long time)
        {
            try
            {
                this.AddPoint(this.points_.Count - 1, x, y, time);
            }
            catch (Exception exp)
            {
                throw new Exception(
                    "[DrawData.AddPoint(x, y, time)]"
                    + "(" + exp.Source + ")\r\n"
                    + "  " + exp.Message,
                    exp);
            }
        }

        public void AlignSelectedObjectsCenteredHorizontal()
        {
            float minX = float.NaN;
            float maxX = float.NaN;
            foreach (PointData pt in this.anchorPoints_.Points)
            {
                if (pt.Selected)
                {
                    if (float.IsNaN(minX) || (minX > pt.X))
                    {
                        minX = pt.X;
                    }
                    if (float.IsNaN(maxX) || (maxX < pt.X))
                    {
                        maxX = pt.X;
                    }
                }
            }

            if (float.IsNaN(minX) || float.IsNaN(maxX))
            {
                return;
            }

            float newX = minX + ((maxX - minX) / 2);
            foreach (PointData pd in this.anchorPoints_.Points)
            {
                if (pd.Selected)
                {
                    pd.X = newX;
                    pd.LinkedPointData.X = newX;
                }
            }
        }

        public void AlignSelectedObjectsCenteredVertical()
        {
            float minY = float.NaN;
            float maxY = float.NaN;
            foreach (PointData pd in this.anchorPoints_.Points)
            {
                if (pd.Selected)
                {
                    if (float.IsNaN(minY) || (minY > pd.Y))
                    {
                        minY = pd.Y;
                    }
                    if (float.IsNaN(maxY) || (maxY < pd.Y))
                    {
                        maxY = pd.Y;
                    }
                }
            }

            if (float.IsNaN(minY) || float.IsNaN(maxY))
            {
                return;
            }

            float newY = minY + ((maxY - minY) / 2);
            foreach (PointData pd in this.anchorPoints_.Points)
            {
                if (pd.Selected)
                {
                    pd.Y = newY;
                    pd.LinkedPointData.Y = newY;
                }
            }
        }

        public void Clear()
        {
            this.points_.Clear();
            this.anchorPoints_.Clear();
            this.backgroundImage_ = null;
            this.penSize_ = 1;
            this.penColor_ = Color.Black;
            this.offsetX_ = 0.0F;
            this.offsetY_ = 0.0F;
            this.scale_ = 1.0F;
            this.text_ = "";
            //this.font_ = SystemFonts.DefaultFont;
            this.font_ = new Font(SystemFonts.DefaultFont.Name, 240F);
            this.textColor_ = Color.LightGray;
        }

        public void CopyTo(
            DrawingData data)
        {
            data.Clear();
            foreach (PointDataCollection iPDC in this.points_)
            {
                PointDataCollection pdc = new PointDataCollection();
                iPDC.CopyTo(pdc);
                data.Points.Add(pdc);
            }
            this.anchorPoints_.CopyTo(data.AnchorPoints);
            data.BackgroundImage = this.backgroundImage_;
            data.PenSize = this.penSize_;
            data.PenColor = this.penColor_;
            data.OffsetX = this.offsetX_;
            data.OffsetY = this.offsetY_;
            data.Scale = this.scale_;
            data.Text = this.text_;
            data.Font = this.font_;
            data.TextColor = this.textColor_;
        }

        public void DeleteSelectedAnchorPoints()
        {
            for (int i = this.anchorPoints_.Points.Count - 1; i >= 0; i--)
            {
                if (this.anchorPoints_.Points[i].Selected)
                {
                    this.anchorPoints_.Points[i].LinkedPointData.Selected = true;
                    this.anchorPoints_.Points.RemoveAt(i);
                }
            }

            for (int i = this.points_.Count - 1; i >= 0; i--)
            {
                this.points_[i].DeleteSelectedPoints();
                if (this.points_[i].Points.Count < 1)
                {
                    this.points_.RemoveAt(i);
                }
            }
        }

        public void DrawAnchorPoints(
            Graphics g)
        {
            try
            {
                for (int i = 0; i < this.anchorPoints_.Points.Count; i++)
                {
                    float x = (this.anchorPoints_.Points[i].X * this.scale_ + this.offsetX_) - ANCHORWIDTH / 2;
                    float y = (this.anchorPoints_.Points[i].Y * this.scale_ + this.offsetY_) - ANCHORHEIGHT / 2;
                    float width = ANCHORWIDTH;
                    float height = ANCHORHEIGHT;
                    Color color = this.anchorPoints_.Points[i].Selected ? Color.Red : Color.White;

                    g.FillRectangle(new SolidBrush(Color.Black), x - 1, y - 1, width + 2, height + 2);
                    g.FillRectangle(new SolidBrush(color), x, y, width, height);
                }
            }
            catch (Exception exp)
            {
                throw new Exception(
                    "[DrawingData.DrawAnchorPoint(g)]"
                    + "(" + exp.Source + ")\r\n"
                    + "  " + exp.Message,
                    exp);
            }
        }

        public void DrawCurve(
            Graphics g,
            Color color,
            float width)
        {
            try
            {
                this.DrawCurve(g, color, width, -1);
            }
            catch (Exception exp)
            {
                throw new Exception(
                    "[DrawData.DrawCurve(g, color, width)]"
                    + "(" + exp.Source + ")\r\n"
                    + "  " + exp.Message,
                    exp);
            }
        }

        public void DrawCurve(
            Graphics g,
            Color color,
            float width,
            long time)
        {
            try
            {
                if (this.points_.Count < 1)
                {
                    return;
                }

                Pen pen = new Pen(color, width * this.scale_);
                pen.StartCap = System.Drawing.Drawing2D.LineCap.Round;
                pen.EndCap = System.Drawing.Drawing2D.LineCap.Round;
                //pen.StartCap = System.Drawing.Drawing2D.LineCap.Square;
                //pen.EndCap = System.Drawing.Drawing2D.LineCap.Square;

                foreach (PointDataCollection ipdc in this.points_)
                {
                    if (ipdc.Points.Count > 1)
                    {
                        PointF[] pts = ipdc.ToPointArray(
                            this.offsetX_,
                            this.offsetY_,
                            this.scale_,
                            time);

                        if (pts == null)
                        {
                            break;
                        }
                        else if (pts.Length < 2)
                        {
                            break;
                        }

                        g.DrawCurve(pen, pts);
                    }
                }
            }
            catch (Exception exp)
            {
                throw new Exception(
                    "[DrawData.DrawCurve(g, color, width, time)]"
                    + "(" + exp.Source + ")\r\n"
                    + "  " + exp.Message,
                    exp);
            }
        }

        public void DrawImage(
            Graphics g)
        {
            if (this.backgroundImage_ == null)
            {
                return;
            }

            try
            {
                g.DrawImage(
                    this.backgroundImage_,
                    this.offsetX_ - this.backgroundImage_.Width * this.scale_ / 2,
                    this.offsetY_ - this.backgroundImage_.Height * this.scale_ / 2,
                    this.backgroundImage_.Width * this.scale_,
                    this.backgroundImage_.Height * this.scale_);
            }
            catch (Exception exp)
            {
                throw new Exception(
                    "[DrawData.DrawImage(g)]"
                    + "(" + exp.Source + ")\r\n"
                    + "  " + exp.Message,
                    exp);
            }
        }

        public void DrawLines(
            Graphics g,
            Color color,
            float width)
        {
            if (this.points_.Count < 1)
            {
                return;
            }

            Pen pen = new Pen(color, width);
            foreach (PointDataCollection pt in this.points_)
            {
                if (pt.Points.Count > 1)
                {
                    g.DrawLines(pen, pt.ToPointArray());
                }
            }
        }

        public void DrawString(
            Graphics g)
        {
            if (this.text_.Length <= 0)
            {
                return;
            }

            StringFormat drawFormat = new StringFormat(StringFormatFlags.MeasureTrailingSpaces | StringFormatFlags.NoWrap);
            System.Windows.Forms.Screen [] screens = System.Windows.Forms.Screen.AllScreens;
            SizeF workingAreaSize = new SizeF();
            foreach (System.Windows.Forms.Screen iscr in screens)
            {
                workingAreaSize.Width += iscr.WorkingArea.Width;
                workingAreaSize.Height += iscr.WorkingArea.Height;
            }
            if (workingAreaSize.Width < 4000f)
            {
                workingAreaSize.Width = 4000f;
            }
            if (workingAreaSize.Height < 4000f)
            {
                workingAreaSize.Height = 4000f;
            }
            SizeF stringSize = this.MeasureString(g, workingAreaSize, drawFormat);
            stringSize.Width *= this.scale_;
            stringSize.Height *= this.scale_;
            Font tmpFont = new Font(this.font_.FontFamily, this.font_.Size * this.scale_);
            g.DrawString(
                this.text_,
                tmpFont,
                new SolidBrush(this.textColor_),
                this.offsetX_ - stringSize.Width / 2,
                this.offsetY_ - stringSize.Height / 2,
                drawFormat);
        }

        public long GetMaxTime()
        {
            long maxTime = 0L;
            if (this.points_.Count > 0)
            {
                foreach (PointDataCollection ipdc in this.points_)
                {
                    long tmpTime = ipdc.GetEndTime();
                    if (tmpTime > maxTime)
                    {
                        maxTime = tmpTime;
                    }
                }
            }
            return maxTime;
        }

        public long GetPointCount()
        {
            long n = 0;
            foreach (PointDataCollection ipdc in this.points_)
            {
                n += ipdc.Points.Count;
            }
            return n;
        }

        public int GetPointsCount(
            int index)
        {
            try
            {
                return this.points_[index].Points.Count;
            }
            catch (Exception exp)
            {
                throw new Exception(
                    "[DrawData.GetPointsCount(index)]"
                    + "(" + exp.Source + ")\r\n"
                    + "  " + exp.Message,
                    exp);
            }
        }

        public RectangleF GetRectangle()
        {
            float left = float.NaN;
            float top = float.NaN;
            float right = float.NaN;
            float bottom = float.NaN;

            foreach (PointDataCollection pdc in this.points_)
            {
                foreach (PointData pd in pdc.Points)
                {
                    if (float.IsNaN(left))
                    {
                        left = pd.X;
                    }
                    else if (left > pd.X)
                    {
                        left = pd.X;
                    }

                    if (float.IsNaN(top))
                    {
                        top = pd.Y;
                    }
                    else if (top > pd.Y)
                    {
                        top = pd.Y;
                    }

                    if (float.IsNaN(right))
                    {
                        right = pd.X;
                    }
                    else if (right < pd.X)
                    {
                        right = pd.X;
                    }

                    if (float.IsNaN(bottom))
                    {
                        bottom = pd.Y;
                    }
                    else if (bottom < pd.Y)
                    {
                        bottom = pd.Y;
                    }
                }
            }

            RectangleF rect = RectangleF.Empty;
            if (!float.IsNaN(left)
                && !float.IsNaN(right)
                && !float.IsNaN(top)
                && !float.IsNaN(bottom))
            {
                rect = new RectangleF(left, top, Math.Abs(right - left), Math.Abs(bottom - top));
            }
            return rect;
        }

        public RectangleF GetRectangleSelected()
        {
            float left = float.NaN;
            float top = float.NaN;
            float right = float.NaN;
            float bottom = float.NaN;

            foreach (PointData pd in this.anchorPoints_.Points)
            {
                pd.LinkedPointData.Selected = pd.Selected;
            }

            foreach (PointDataCollection pdc in this.points_)
            {
                foreach (PointData pd in pdc.Points)
                {
                    if (!pd.Selected)
                    {
                        continue;
                    }

                    if (float.IsNaN(left))
                    {
                        left = pd.X;
                    }
                    else if (left > pd.X)
                    {
                        left = pd.X;
                    }

                    if (float.IsNaN(top))
                    {
                        top = pd.Y;
                    }
                    else if (top > pd.Y)
                    {
                        top = pd.Y;
                    }

                    if (float.IsNaN(right))
                    {
                        right = pd.X;
                    }
                    else if (right < pd.X)
                    {
                        right = pd.X;
                    }

                    if (float.IsNaN(bottom))
                    {
                        bottom = pd.Y;
                    }
                    else if (bottom < pd.Y)
                    {
                        bottom = pd.Y;
                    }
                }
            }

            RectangleF rect = RectangleF.Empty;
            if (!float.IsNaN(left)
                && !float.IsNaN(right)
                && !float.IsNaN(top)
                && !float.IsNaN(bottom))
            {
                rect = new RectangleF(left, top, Math.Abs(right - left), Math.Abs(bottom - top));
            }
            return rect;
        }

        public bool IsAnchorExist(
            float x,
            float y)
        {
            PointF pt = this.NormalizePoint(x, y);

            for (int i = this.anchorPoints_.Points.Count - 1; i >= 0; i--)
            {
                float x0 = this.anchorPoints_.Points[i].X - ANCHORWIDTH / 2 / this.scale_;
                float y0 = this.anchorPoints_.Points[i].Y - ANCHORHEIGHT / 2 / this.scale_;
                float x1 = this.anchorPoints_.Points[i].X + ANCHORWIDTH / 2 / this.scale_;
                float y1 = this.anchorPoints_.Points[i].Y + ANCHORHEIGHT / 2 / this.scale_;

                if ((x0 <= pt.X) && (x1 >= pt.X)
                    && (y0 <= pt.Y) && (y1 >= pt.Y))
                {
                    return true;
                }
            }
            return false;
        }

        public bool IsAnchorSelected(
            float x,
            float y)
        {
            PointF pt = this.NormalizePoint(x, y);

            for (int i = this.anchorPoints_.Points.Count - 1; i >= 0; i--)
            {
                if (this.anchorPoints_.Points[i].Selected)
                {
                    float x0 = this.anchorPoints_.Points[i].X - ANCHORWIDTH / 2 / this.scale_;
                    float y0 = this.anchorPoints_.Points[i].Y - ANCHORHEIGHT / 2 / this.scale_;
                    float x1 = this.anchorPoints_.Points[i].X + ANCHORWIDTH / 2 / this.scale_;
                    float y1 = this.anchorPoints_.Points[i].Y + ANCHORHEIGHT / 2 / this.scale_;

                    if ((x0 <= pt.X) && (x1 >= pt.X)
                        && (y0 <= pt.Y) && (y1 >= pt.Y))
                    {
                        return true;
                    }
                }
            }
            return false;
        }

        public bool IsSelectedAnchorExist()
        {
            foreach (PointData ianchor in this.anchorPoints_.Points)
            {
                if (ianchor.Selected)
                {
                    return true;
                }
            }
            return false;
        }

        public SizeF MeasureString(
            Graphics g,
            SizeF layoutArea,
            StringFormat stringFormat)
        {
            return g.MeasureString(
                this.text_,
                this.font_,
                layoutArea,
                stringFormat);
        }

        public void MoveSelectedPointTo(
            float offsetX,
            float offsetY,
            bool unselect)
        {
            foreach (PointData pd in this.anchorPoints_.Points)
            {
                if (pd.Selected)
                {
                    pd.MoveTo(offsetX / this.scale_, offsetY / this.scale_);
                }
                pd.LinkedPointData.Selected = pd.Selected;
            }

            foreach (PointDataCollection pdc in this.points_)
            {
                pdc.MoveSelectedPoitTo(offsetX / this.scale_, offsetY / this.scale_);
            }

            if (unselect)
            {
                this.UnselectPoints();
            }
        }

        public void MoveSelectedPointTo(
            float offsetX,
            float offsetY)
        {
            this.MoveSelectedPointTo(offsetX, offsetY, true);
        }

        public void MoveTo(
            float offsetX,
            float offsetY)
        {
            foreach (PointDataCollection pdc in this.points_)
            {
                pdc.MoveTo(offsetX, offsetY);
            }
        }

        public PointF NormalizePoint(
            float x,
            float y)
        {
            PointF pt = new PointF();
            pt.X = (x - this.offsetX_) / this.scale_;
            pt.Y = (y - this.offsetY_) / this.scale_;
            return pt;
        }

        public void ReadData(
            string fileName)
        {
            try
            {
                this.Clear();

                using (BinaryReader br = new BinaryReader(
                    File.Open(fileName, FileMode.Open)))
                {
                    string tag = "";
                    int pointsCount = 0;
                    int pointDataCount = 0;
                    int width = 0;
                    int height = 0;
                    string fontName = "";
                    float fontSize = 0.0F;
                    int textColorARGB = 0;

                    while (br.BaseStream.Position < br.BaseStream.Length)
                    {
                        tag = br.ReadString();
                        switch (tag)
                        {
                            case "DrawingData":
                                break;
                            case "PenSize":
                                this.penSize_ = br.ReadInt32();
                                break;
                            case "PenColor":
                                int penARGB = br.ReadInt32();
                                this.penColor_ = Color.FromArgb(penARGB);
                                break;
                            case "PointDataCollection":
                                pointsCount = br.ReadInt32();
                                for (int i = 0; i < pointsCount; i++)
                                {
                                    PointDataCollection pdc = new PointDataCollection();

                                    //  PointData
                                    tag = br.ReadString();
                                    switch (tag)
                                    {
                                        case "PointData":
                                            pointDataCount = br.ReadInt32();
                                            for (int j = 0; j < pointDataCount; j++)
                                            {
                                                PointData pd = new PointData();
                                                pd.X = br.ReadSingle();
                                                pd.Y = br.ReadSingle();
                                                pd.Time = br.ReadInt64();

                                                pdc.Points.Add(pd);
                                            }

                                            if (pdc.Points.Count > 0)
                                            {
                                                this.points_.Add(pdc);
                                            }
                                            break;
                                    }
                                }
                                break;
                            case "Bitmap":
                                width = br.ReadInt32();
                                height = br.ReadInt32();
                                if (width * height > 0)
                                {
                                    backgroundImage_ = new Bitmap(width, height);
                                    for (int y = 0; y < height; y++)
                                    {
                                        for (int x = 0; x < width; x++)
                                        {
                                            int bitmapARGB = br.ReadInt32();
                                            backgroundImage_.SetPixel(x, y, Color.FromArgb(bitmapARGB));
                                        }
                                    }
                                }
                                break;
                            case "FontName":
                                fontName = br.ReadString();
                                if ((fontName.Length > 0) && (fontSize > 0))
                                {
                                    this.font_ = new Font(fontName, fontSize);
                                }
                                break;
                            case "FontSize":
                                fontSize = br.ReadSingle();
                                if ((fontName.Length > 0) && (fontSize > 0))
                                {
                                    this.font_ = new Font(fontName, fontSize);
                                }
                                break;
                            case "TextColor":
                                textColorARGB = br.ReadInt32();
                                this.textColor_ = Color.FromArgb(textColorARGB);
                                break;
                            case "Text":
                                this.text_ = br.ReadString();
                                break;
                        }
                    }
                }
            }
            catch (Exception exp)
            {
                throw new Exception(
                    "[DrawData.ReadData(fileName)]"
                    + "(" + exp.Source + ")\r\n"
                    + "  " + exp.Message,
                    exp);
            }
        }

        public void RemoveAt(
            int index)
        {
            try
            {
                this.points_.RemoveAt(index);
            }
            catch (Exception exp)
            {
                throw new Exception(
                    "[DrawData.RemoveAt(index)]"
                    + "(" + exp.Source + ")\r\n"
                    + "  " + exp.Message,
                    exp);
            }
        }

        public void ScaleObject(
            float scaleX,
            float scaleY)
        {
            RectangleF rect = this.GetRectangle();
            float offsetX = rect.Left + rect.Width / 2;
            float offsetY = rect.Top + rect.Height / 2;

            foreach (PointDataCollection pdc in this.points_)
            {
                pdc.Scale(offsetX, offsetY, scaleX, scaleY);
            }
        }

        public PointF ScalePoint(
            float x,
            float y,
            float scale)
        {
            PointF pt = new PointF();
            pt.X = scale * x - this.offsetX_ * scale + this.offsetX_;
            pt.Y = scale * y - this.offsetY_ * scale + this.offsetY_;
            return pt;
        }

        public int SelectAnchor(
            float x,
            float y)
        {
            PointF pt = this.NormalizePoint(x, y);

            for (int i = this.anchorPoints_.Points.Count - 1; i >= 0; i--)
            {
                float x0 = this.anchorPoints_.Points[i].X - ANCHORWIDTH / 2 / this.scale_;
                float y0 = this.anchorPoints_.Points[i].Y - ANCHORHEIGHT / 2 / this.scale_;
                float x1 = this.anchorPoints_.Points[i].X + ANCHORWIDTH / 2 / this.scale_;
                float y1 = this.anchorPoints_.Points[i].Y + ANCHORHEIGHT / 2 / this.scale_;

                if ((x0 <= pt.X) && (x1 >= pt.X)
                    && (y0 <= pt.Y) && (y1 >= pt.Y))
                {
                    //Console.WriteLine("[anchor]({0}, {1}) -> ({2}, {3}):({4}, {5}) - ({6}, {7})", x, y, pt.X, pt.Y, x0, y0, x1, y1);

                    this.anchorPoints_.Points[i].Selected = true;
                    return i;
                }
            }
            return -1;
        }

        public void SelectAnchorAll()
        {
            foreach (PointData ipd in this.anchorPoints_.Points)
            {
                ipd.Selected = true;
            }
        }

        public long SelectAnchorInRegion(
            float left,
            float top,
            float right,
            float bottom)
        {
            PointF pt0 = this.NormalizePoint(left, top);
            PointF pt1 = this.NormalizePoint(right, bottom);
            long n = 0;
            foreach (PointData ipd in this.anchorPoints_.Points)
            {
                if ((ipd.X <= pt1.X) && (ipd.X >= pt0.X)
                    && (ipd.Y <= pt1.Y) && (ipd.Y >= pt0.Y))
                {
                    ipd.Selected = true;
                    ++n;
                }
            }
            return n;
        }

        public void SetAnchorPoints()
        {
            this.anchorPoints_.Clear();
            for (int i = 0; i < this.points_.Count; i++)
            {
                for (int j = 0; j < this.points_[i].Points.Count; j++)
                {
                    PointData pd = new PointData(
                        this.points_[i].Points[j].X,
                        this.points_[i].Points[j].Y,
                        this.points_[i].Points[j].Time);
                    pd.LinkedPointData = this.points_[i].Points[j];
                    this.anchorPoints_.Add(pd);
                }
            }
        }

        public void SetStartTime(
            long time)
        {
            long timeDiff = 0;

            foreach (PointDataCollection pdc in this.points_)
            {
                bool p = false;
                foreach (PointData pt in pdc.Points)
                {
                    p = true;
                    timeDiff = time - pt.Time;
                    break;
                }
                if (p)
                {
                    break;
                }
            }

            if (timeDiff == 0)
            {
                return;
            }

            foreach (PointDataCollection pdc in this.points_)
            {
                foreach (PointData pt in pdc.Points)
                {
                    pt.Time += timeDiff;
                }
            }
        }

        public void SetStrokeInterval(
            long strokeTimeInterval)
        {
            long tmpTime = 0L;
            for (int i = 0; i < this.points_.Count; i++)
            {
                PointDataCollection ipdc = this.points_[i];
                if (i == 0)
                {
                    tmpTime = ipdc.GetEndTime();
                }
                else
                {
                    ipdc.ResetTime(tmpTime + strokeTimeInterval);
                    tmpTime = ipdc.GetEndTime();
                }
            }
        }

        public PointF[] ToPointArray()
        {
            PointF[] pts = null;
            foreach (PointDataCollection pdc in this.points_)
            {
                if (pdc.Points.Count > 0)
                {
                    PointF[] tmpPts = pdc.ToPointArray();
                    int index = 0;
                    if (pts == null)
                    {
                        pts = new PointF[tmpPts.Length];
                    }
                    else
                    {
                        index = pts.Length;
                        Array.Resize<PointF>(ref pts, pts.Length + tmpPts.Length);
                    }
                    tmpPts.CopyTo(pts, index);
                }
            }
            return pts;
        }

        public void TransformToEllipse(
            RectangleF rect)
        {
            RectangleF selRect = this.GetRectangleSelected();
            RectangleF toRect = new RectangleF(
                this.NormalizePoint(rect.X, rect.Y),
                new SizeF(rect.Width / this.scale_, rect.Height / this.scale_));
            this.MoveSelectedPointTo(
                -selRect.Left * this.scale_ - (selRect.Width * this.scale_ / 2), 
                -selRect.Top * this.scale_ - (selRect.Height * this.scale_ / 2),
                false);
            double width = toRect.Width / 2;
            double height = toRect.Height / 2;
            foreach (PointDataCollection ipdc in this.points_)
            {
                foreach (PointData ipd in ipdc.Points)
                {
                    if (ipd.Selected)
                    {
                        double a = Math.Atan2((double)ipd.Y, (double)ipd.X);
                        if ((ipd.X < 0) && (ipd.Y < 0))
                        {
                            a = Math.Atan2(-ipd.Y, -ipd.X) + Math.PI;
                        }
                        else if ((ipd.X >= 0) && (ipd.Y < 0))
                        {
                            a = 2 * Math.PI + Math.Atan2(ipd.Y, ipd.X);
                        }
                        float x = (float)(width * Math.Cos(a));
                        float y = (float)(height * Math.Sin(a));
                        //Console.WriteLine("({0}, {1}) -> ({2}, {3}) {4}", ipd.X, ipd.Y, x, y, a * 180 / Math.PI);
                        ipd.X = x + toRect.Left + toRect.Width / 2;
                        ipd.Y = y + toRect.Top + toRect.Height / 2;
                    }
                }
            }
            //this.MoveSelectedPointTo(
            //    (toRect.Left / this.scale_ + toRect.Width / this.scale_ / 2),
            //    (toRect.Top / this.scale_ + toRect.Height / this.scale_ / 2),
            //    false);
        }

        public int UnselectAnchor(
            float x,
            float y)
        {
            PointF pt = this.NormalizePoint(x, y);

            for (int i = this.anchorPoints_.Points.Count - 1; i >= 0; i--)
            {
                float x0 = this.anchorPoints_.Points[i].X - ANCHORWIDTH / 2 / this.scale_;
                float y0 = this.anchorPoints_.Points[i].Y - ANCHORHEIGHT / 2 / this.scale_;
                float x1 = this.anchorPoints_.Points[i].X + ANCHORWIDTH / 2 / this.scale_;
                float y1 = this.anchorPoints_.Points[i].Y + ANCHORHEIGHT / 2 / this.scale_;

                if ((x0 <= pt.X) && (x1 >= pt.X)
                    && (y0 <= pt.Y) && (y1 >= pt.Y))
                {
                    //Console.WriteLine("[anchor]({0}, {1}) -> ({2}, {3}):({4}, {5}) - ({6}, {7})", x, y, pt.X, pt.Y, x0, y0, x1, y1);

                    this.anchorPoints_.Points[i].Selected = false;
                    return i;
                }
            }
            return -1;
        }

        public void UnselectAnchorAll()
        {
            foreach (PointData pd in this.anchorPoints_.Points)
            {
                pd.Selected = false;
            }
            this.UnselectPoints();
        }

        public void UnselectPoints()
        {
            foreach (PointDataCollection pdc in this.points_)
            {
                pdc.UnselectPoints();
            }
        }

        public void WriteData(
            string fileName)
        {
            try
            {
                using (BinaryWriter bw = new BinaryWriter(
                    File.Open(fileName, FileMode.Create)))
                {
                    bw.Write("DrawingData");
                    bw.Write("PenSize");
                    bw.Write(this.penSize_);
                    bw.Write("PenColor");
                    bw.Write(this.penColor_.ToArgb());
                    bw.Write("PointDataCollection");
                    bw.Write(this.points_.Count);
                    for (int i = 0; i < this.points_.Count; i++)
                    {
                        bw.Write("PointData");
                        bw.Write(this.points_[i].Points.Count);
                        foreach (PointData pd in this.points_[i].Points)
                        {
                            bw.Write(pd.X);
                            bw.Write(pd.Y);
                            bw.Write(pd.Time);
                        }
                    }
                    bw.Write("Bitmap");
                    if (this.backgroundImage_ == null)
                    {
                        bw.Write(0);
                        bw.Write(0);
                    }
                    else
                    {
                        bw.Write(this.backgroundImage_.Width);
                        bw.Write(this.backgroundImage_.Height);
                        if ((this.backgroundImage_.Width * this.backgroundImage_.Height) > 0)
                        {
                            for (int y = 0; y < this.backgroundImage_.Height; y++)
                            {
                                for (int x = 0; x < this.backgroundImage_.Width; x++)
                                {
                                    bw.Write(this.backgroundImage_.GetPixel(x, y).ToArgb());
                                }
                            }
                        }
                    }
                    bw.Write("FontName");
                    bw.Write(this.font_.Name);
                    bw.Write("FontSize");
                    bw.Write(this.font_.Size);
                    bw.Write("TextColor");
                    bw.Write(this.textColor_.ToArgb());
                    bw.Write("Text");
                    bw.Write(this.text_);
                }
            }
            catch (Exception exp)
            {
                throw new Exception(
                    "[DrawData.WriteData(fileName)]"
                    + "(" + exp.Source + ")\r\n"
                    + "  " + exp.Message,
                    exp);
            }
        }

        public void WriteImage(
            string fileNameWithFullPath,
            int width,
            int height,
            bool textDraw,
            bool backgroundImageDraw,
            ImageFormat imageFormat)
        {
            try
            {
                if ((imageFormat == ImageFormat.Emf)
                    || (imageFormat == ImageFormat.Wmf))
                {
                    this.WriteMetaFile(
                        fileNameWithFullPath,
                        width,
                        height,
                        textDraw,
                        backgroundImageDraw,
                        imageFormat);
                    return;
                }

                Bitmap canvas = new Bitmap(width, height);
                Graphics g = Graphics.FromImage(canvas);
                g.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.HighQuality;
                g.PageUnit = GraphicsUnit.Pixel;
                g.Clear(Color.White);
                this.OffsetX = canvas.Width / 2;
                this.OffsetY = canvas.Height / 2;
                if (backgroundImageDraw)
                {
                    this.DrawImage(g);
                }
                if (textDraw)
                {
                    this.DrawString(g);
                }
                this.DrawCurve(
                    g,
                    this.penColor_,
                    this.penSize_);
                g.Dispose();

                canvas.Save(
                    fileNameWithFullPath,
                    imageFormat);

                canvas.Dispose();
            }
            catch (Exception exp)
            {
                throw new Exception(
                    "[DrawData.WriteImage(fileNameWithFullPath, width, height \r\n, textDraw, backgroundImageDraw, imageFormat)]"
                    + "(" + exp.Source + ")\r\n"
                    + "  " + exp.Message,
                    exp);
            }
        }

        public long WriteImages(
            string fileNameWithFullPath,
            float fps,
            int width,
            int height,
            bool textDraw,
            bool backgroundImageDraw,
            ImageFormat imageFormat)
        {
            string directoryName = fileNameWithFullPath;
            if (directoryName.LastIndexOf(".") > 0)
            {
                directoryName = directoryName.Substring(0, directoryName.LastIndexOf("."));
            }
            else
            {
                return -1;
            }

            try
            {
                return WriteImages(
                    fileNameWithFullPath,
                    fps,
                    width,
                    height,
                    textDraw,
                    backgroundImageDraw,
                    imageFormat,
                    directoryName);
            }
            catch (Exception exp)
            {
                throw new Exception(
                    "[DrawData.WriteImages(fileNameWithFullPath, fps, width, height \r\n, textDraw, backgroundImageDraw, imageFormat)]"
                    + "(" + exp.Source + ")\r\n"
                    + "  " + exp.Message,
                    exp);
            }
        }

        public long WriteImages(
            string fileNameWithFullPath,
            float fps,
            int width,
            int height,
            bool textDraw,
            bool backgroundImageDraw,
            ImageFormat imageFormat,
            string folderPathToWrite)
        {
            bool directoryCreated = false;
            string directoryName = folderPathToWrite;
            if (!Directory.Exists(directoryName))
            {
                Directory.CreateDirectory(directoryName);
                directoryCreated = true;
            }

            string fileName = fileNameWithFullPath;
            if (fileName.LastIndexOf(@"\") > 0)
            {
                if (fileName.LastIndexOf(".") <= 0)
                {
                    fileName = fileName.Substring(
                        fileName.LastIndexOf(@"\") + 1,
                        0);
                }
                else
                {
                    fileName = fileName.Substring(
                        fileName.LastIndexOf(@"\") + 1,
                        fileName.LastIndexOf(".") - fileName.LastIndexOf(@"\") - 1);
                }
                if (fileName.Length > 0)
                {
                    fileName += "_";
                }
            }
            else
            {
                if (directoryCreated)
                {
                    Directory.Delete(directoryName);
                }
                return -1;
            }

            string extension = ".img";
            if (imageFormat == ImageFormat.Bmp)
            {
                extension = ".bmp";
            }
            else if (imageFormat == ImageFormat.Emf)
            {
                extension = ".emf";
            }
            else if (imageFormat == ImageFormat.Gif)
            {
                extension = ".gif";
            }
            else if (imageFormat == ImageFormat.Jpeg)
            {
                extension = ".jpg";
            }
            else if (imageFormat == ImageFormat.Png)
            {
                extension = ".png";
            }

            try
            {
                float frameTime = (float)(1000 / fps);      //  milli second
                long maxTime = this.GetMaxTime();           //  ticks(100 nano second(= 1/10000 milli second)
                long elapsedTime_ = 0;
                int cnt = 0;
                while (elapsedTime_ <= maxTime)
                {
                    Bitmap canvas = new Bitmap(width, height);
                    Graphics g = Graphics.FromImage(canvas);
                    g.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.HighQuality;
                    g.PageUnit = GraphicsUnit.Pixel;
                    g.Clear(Color.White);
                    this.OffsetX = canvas.Width / 2;
                    this.OffsetY = canvas.Height / 2;
                    if (backgroundImageDraw)
                    {
                        this.DrawImage(g);
                    }
                    if (textDraw)
                    {
                        this.DrawString(g);
                    }
                    DrawCurve(
                        g,
                        this.penColor_,
                        this.penSize_,
                        elapsedTime_);
                    g.Dispose();

                    ++cnt;

                    canvas.Save(
                        directoryName + @"\" + fileName + cnt.ToString("0000") + extension,
                        imageFormat);

                    canvas.Dispose();

                    if (elapsedTime_ < maxTime)
                    {
                        elapsedTime_ += (long)(frameTime * 10000);      //  nano second
                        if (elapsedTime_ > maxTime)
                        {
                            elapsedTime_ = maxTime;
                        }
                    }
                    else
                    {
                        break;
                    }
                }

                return cnt;
            }
            catch (Exception exp)
            {
                throw new Exception(
                    "[DrawData.WriteImages(fileNameWithFullPath, fps, width, height\r\n"
                    + ", textDraw, backgroundImageDraw, imageFormat\r\n"
                    + ", folderPathToWrite)]"
                    + "(" + exp.Source + ")\r\n"
                    + "  " + exp.Message,
                    exp);
            }
        }

        public void WriteMetaFile(
            string fileNameWithFullPath,
            int width,
            int height,
            bool textDraw,
            bool backgroundImageDraw,
            ImageFormat imageFormat)
        {
            try
            {
                float pensize = this.penSize_ + 0.5f;
                RectangleF tmpRect = this.GetRectangle();
                RectangleF rect = new RectangleF(
                    tmpRect.X - this.penSize_ / 2,
                    tmpRect.Y - this.penSize_ / 2,
                    tmpRect.Width + (float)(this.penSize_ * 1.1),
                    tmpRect.Height + (float)(this.penSize_ * 1.1));
                Bitmap canvas = new Bitmap((int)rect.Width, (int)rect.Height);
                canvas.MakeTransparent(Color.White);
                Graphics tmpG = Graphics.FromImage(canvas);
                tmpG.Clear(Color.White);

                IntPtr hDC = tmpG.GetHdc();
                //MemoryStream memStream = new MemoryStream();
                //Metafile metaFile = new Metafile(memStream, hDC, EmfType.EmfPlusDual);
                Metafile metaFile = new Metafile(
                    fileNameWithFullPath, 
                    hDC, 
                    rect,
                    MetafileFrameUnit.Pixel,
                    EmfType.EmfPlusDual);
                tmpG.ReleaseHdc();

                Graphics g = Graphics.FromImage(metaFile);
                g.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.HighQuality;
                g.PageUnit = GraphicsUnit.Pixel;
                //g.Clear(Color.White);

                this.OffsetX = 0;   // rect.Width / 2;
                this.OffsetY = 0;   // rect.Height / 2;
                if (backgroundImageDraw)
                {
                    this.DrawImage(g);
                }
                if (textDraw)
                {
                    this.DrawString(g);
                }
                this.DrawCurve(
                    g,
                    this.penColor_,
                    this.penSize_);

                //FileStream fileStream = new FileStream(
                //    fileNameWithFullPath,
                //    FileMode.Create,
                //    FileAccess.Write);
                //fileStream.Write(memStream.GetBuffer(), 0, (int)memStream.Length);

                //fileStream.Close();
                //memStream.Close();

                g.Dispose();
                metaFile.Dispose();
            }
            catch (Exception exp)
            {
                throw new Exception(
                    "[DrawData.WriteMetaFile(fileNameWithFullPath, width, height \r\n, textDraw, backgroundImageDraw, imageFormat)]"
                    + "(" + exp.Source + ")\r\n"
                    + "  " + exp.Message,
                    exp);
            }
        }

        #endregion  //  method
    }
}
