﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Text;

namespace OsakaUniv.RIWL.Koudo.Writimation
{
    public class PointDataCollection
    {
        public const float CURVETHRESHOLD = 0.995F;
        private List<PointData> points_;

        #region constructor
        //
        //  constructor
        //

        public PointDataCollection()
        {
            points_ = new List<PointData>();
        }

        public PointDataCollection(
            float x,
            float y,
            long time)
        {
            points_ = new List<PointData>();
            this.Add(x, y, time);
        }

        public PointDataCollection(
            PointF pt,
            long time)
        {
            this.points_ = new List<PointData>();
            this.Add(pt.X, pt.Y, time);
        }

        #endregion  //  constructor

        #region property
        //
        //  property
        //

        public List<PointData> Points
        {
            get { return this.points_; }
            set { this.points_ = value; }
        }

        #endregion  //  property

        #region method
        //
        //  method
        //

        public void Add(
            PointData item)
        {
            try
            {
                this.points_.Add(item);

                if (this.points_.Count > 1)
                {
                    if (this.points_[this.points_.Count - 1].ComparePointTo(
                        this.points_[this.points_.Count - 2]))
                    {
                        this.points_.RemoveAt(this.points_.Count - 1);
                    }
                }
                else if (this.points_.Count >= 3)
                {
                    float r = CalculateCurvature(
                        this.points_[this.points_.Count - 3],
                        this.points_[this.points_.Count - 2],
                        this.points_[this.points_.Count - 1]);

                    //Console.WriteLine("{0}", r);

                    if (r > CURVETHRESHOLD)
                    {
                        this.RemoveAt(this.points_.Count - 2);
                    }
                }
            }
            catch (Exception exp)
            {
                throw new Exception(
                    "[PointDataCollection.Add(item)] "
                    + "(" + exp.Source + ")" + "\r\n"
                    + "  " + exp.Message,
                    exp);
            }
        }

        public void Add(
            float x,
            float y,
            long time)
        {
            try
            {
                this.Add(new PointData(x, y, time));
            }
            catch (Exception exp)
            {
                throw new Exception(
                    "[PointDataCollection.Add(x, y, time)] "
                    + "(" + exp.Source + ")" + "\r\n"
                    + "  " + exp.Message,
                    exp);
            }
        }

        public void Add(
            PointF pt,
            long time)
        {
            try
            {
                this.Add(new PointData(pt.X, pt.Y, time));
            }
            catch (Exception exp)
            {
                throw new Exception(
                    "[PointDataCollection.Add(x, y, time)] "
                    + "(" + exp.Source + ")" + "\r\n"
                    + "  " + exp.Message,
                    exp);
            }
        }

        /// <summary>
        /// caluculate curvature
        /// </summary>
        /// <remarks>
        /// formulas.
        /// 
        ///                        (x1 - x0) * (x2 - x1) + (y1 - y0) * (y2 - y1)
        /// R = cos(T) = -----------------------------------------------------------------
        ///              Sqrt((x1 - x0)^2 + (y1 - y0)^2) * Sqrt((x2 - x1)^2 + (y2 - y1)^2)
        /// 
        /// T : theta = angle
        /// </remarks>
        /// <param name="pt0">first point data</param>
        /// <param name="pt1">second point data</param>
        /// <param name="pt2">third point data</param>
        /// <returns>angle</returns>
        public float CalculateCurvature(
            PointData pt0,
            PointData pt1,
            PointData pt2)
        {
            try
            {
                float r = (float)(((pt1.X - pt0.X) * (pt2.X - pt1.X) + (pt1.Y - pt0.Y) * (pt2.Y - pt1.Y))
                    / (Math.Sqrt((pt1.X - pt0.X) * (pt1.X - pt0.X) + (pt1.Y - pt0.Y) * (pt1.Y - pt0.Y))
                        * Math.Sqrt((pt2.X - pt1.X) * (pt2.X - pt1.X) + (pt2.Y - pt1.Y) * (pt2.Y - pt1.Y))));
                return r;
            }
            catch (Exception exp)
            {
                throw new Exception(
                    "[PointDataCollection.CalculateCurvature(pt0, pt1, pt2)] "
                    + "(" + exp.Source + ")" + "\r\n"
                    + "  " + exp.Message,
                    exp);
            }
        }

        public void Clear()
        {
            this.points_.Clear();
        }

        public void CopyTo(
            PointDataCollection data)
        {
            data.Clear();
            foreach (PointData pd in this.points_)
            {
                PointData tmpPd = new PointData();
                pd.CopyTo(tmpPd);
                data.Points.Add(tmpPd);
            }
        }

        public void DeleteSelectedPoints()
        {
            for (int i = this.points_.Count - 1; i >= 0; i--)
            {
                if (this.points_[i].Selected)
                {
                    this.points_.RemoveAt(i);
                }
            }
        }

        public long GetEndTime()
        {
            long time = 0;
            if (this.points_.Count > 0)
            {
                foreach (PointData ipd in this.points_)
                {
                    time = (time < ipd.Time) ? ipd.Time : time;
                }
            }
            return time;
        }

        public long GetStartTime()
        {
            long time = -1;
            if (this.points_.Count > 0)
            {
                foreach (PointData ipd in this.points_)
                {
                    time = ((time < 0) || (time > ipd.Time)) ? ipd.Time : time;
                }
            }
            return time;
        }

        public void MoveSelectedPoitTo(
            float offsetX,
            float offsetY)
        {
            foreach (PointData pd in this.points_)
            {
                if (pd.Selected)
                {
                    pd.MoveTo(offsetX, offsetY);
                }
            }
        }

        public void MoveTo(
            float offsetX,
            float offsetY)
        {
            foreach (PointData pd in this.points_)
            {
                pd.MoveTo(offsetX, offsetY);
            }
        }

        public void RemoveAt(
            int index)
        {
            try
            {
                this.points_.RemoveAt(index);
            }
            catch (Exception exp)
            {
                throw new Exception(
                    "[PointDataCollection.RemoveAt] "
                    + "(" + exp.Source + ")" + "\r\n"
                    + "  " + exp.Message,
                    exp);
            }
        }

        public void ResetTime(
            long newTime)
        {
            if (this.points_.Count <= 0)
            {
                return;
            }

            long timediff = newTime - this.GetStartTime();
            foreach (PointData ipt in this.points_)
            {
                ipt.Time += timediff;
            }
        }

        /// <summary>
        /// scaling points by affine transformation.
        /// </summary>
        /// <remarks>
        /// 1. moving to coordinates that center on the origin.
        /// 2. scale
        /// 3. moving to original coordinates.
        /// 
        /// formulas
        ///
        ///                     [ 1   0   0 ][ Sx  0   0 ][ 1   0   0 ]
        /// [X' Y' W'] = [X Y 1][ 0   1   0 ][ 0   Sy  0 ][ 0   1   0 ]
        ///                     [-Px -Py  1 ][ 0   0   1 ][ Px  Py  1 ]
        /// 
        /// X' = SxX - PxSx + Px
        /// Y' = SyY - PySy + Py
        /// W' = 1
        /// </remarks>
        public void Scale(
            float offsetX,
            float offsetY,
            float scaleX,
            float scaleY)
        {
            foreach (PointData pd in this.points_)
            {
                pd.X = scaleX * pd.X - offsetX * scaleX + offsetX;
                pd.Y = scaleY * pd.Y - offsetY * scaleY + offsetY;
            }
        }

        public PointF[] ToPointArray()
        {
            return this.ToPointArray(0, 0, 1, -1);
        }

        public PointF[] ToPointArray(
            long time)
        {
            return this.ToPointArray(0, 0, 1, time);
        }

        public PointF[] ToPointArray(
            float offsetX,
            float offsetY,
            float scale,
            long time)
        {
            PointF[] pts = new PointF[this.points_.Count];
            int n = 0;
            for (int i = 0; i < this.points_.Count; i++)
            {
                if ((time >= 0) && (this.points_[i].Time > time))
                {
                    break;
                }

                pts[i].X = this.points_[i].X * scale + offsetX;
                pts[i].Y = this.points_[i].Y * scale + offsetY;
                n++;
            }

            if (n == 0)
            {
                pts = null;
            }
            else
            {
                Array.Resize<PointF>(ref pts, n);
            }

            return pts;
        }

        public void UnselectPoints()
        {
            foreach (PointData pd in this.points_)
            {
                pd.Selected = false;
            }
        }

        #endregion  //  method
    }
}
